package de.landsh.opendata.csw2dcat;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.apache.hc.client5.http.classic.methods.HttpPost;
import org.apache.hc.client5.http.impl.classic.CloseableHttpClient;
import org.apache.hc.client5.http.impl.classic.CloseableHttpResponse;
import org.apache.hc.client5.http.impl.classic.HttpClients;
import org.apache.hc.core5.http.HttpHeaders;
import org.apache.hc.core5.http.io.entity.StringEntity;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class CswInterface {

    public static final int RECORDS_PER_PAGE = 50;

    private static final Logger log = LoggerFactory.getLogger(CswInterface.class);
    private final SAXReader saxReader = new SAXReader();
    private final String url;
    /**
     * List of item types (i.e. dataset, datasetCollection, series, application, tile, service) that should <b>not</b> be searched.
     */
    private String[] typeExclude;
    /**
     * List of item types (i.e. dataset, datasetCollection, series, application, tile, service) that should be searched.
     * If not set all item types will be searched.
     */
    private String[] typeInclude;
    private CloseableHttpClient httpClient = HttpClients.createMinimal();
    private String dumpCswResponse = null;
    /**
     * Name of a directory. If set all incoming messages will be logged into files in this directory.
     */
    private String verboseLogging = null;
    private boolean distributedSearch = false;
    private boolean filterOpendata = false;
    /**
     * Ask the CSW server in the GetRecords operation to sort the results by Identifier
     */
    private boolean sortResults = true;
    private String authorization = null;
    public CswInterface(String url) {
        this.url = url;
    }

    public CswInterface(BridgeSettings settings) {
        this.url = settings.cswURL;
        setDistributedSearch(settings.distributedSearch);
        filterOpendata = settings.filterOpenData;
        authorization = settings.authorization;
        sortResults = settings.sortResults;
        typeExclude = settings.typeExclude;
        typeInclude = settings.typeInclude;
    }

    public void setVerboseLogging(String verboseLogging) {
        this.verboseLogging = verboseLogging;
    }

    public void setTypeExclude(String[] typeExclude) {
        this.typeExclude = typeExclude;
    }

    public void setTypeInclude(String[] typeInclude) {
        this.typeInclude = typeInclude;
    }

    public void setSortResults(boolean sortResults) {
        this.sortResults = sortResults;
    }

    boolean hasTypeInclude() {
        return typeInclude != null && typeInclude.length > 0;
    }

    boolean hasTypeExclude() {
        return typeExclude != null && typeExclude.length > 0;
    }

    boolean hasMultipleTypeExclude() {
        return typeExclude != null && typeExclude.length > 1;
    }

    void setHttpClient(CloseableHttpClient httpClient) {
        this.httpClient = httpClient;
    }

    public boolean isFilterOpendata() {
        return filterOpendata;
    }

    public void setFilterOpendata(boolean filterOpendata) {
        this.filterOpendata = filterOpendata;
    }

    public void setDistributedSearch(boolean distributedSearch) {
        this.distributedSearch = distributedSearch;
    }

    public int getRecordsPerPage() {
        return RECORDS_PER_PAGE;
    }

    public void setDumpCswResponse(String s) {
        this.dumpCswResponse = s;
    }

    /**
     * @param startPosition number of the first requested record. Used for paging. Will be sent to the WFS server as
     *                      <code>startPosition</code> First record is 1.
     * @param status        if a {@link GetRecordsStatus} object is specified, it will be filled with the status
     *                      (numberOfRecordsMatched, numberOfRecordsReturned) of the getRecords operation
     * @return list of <code>gmd:MD_Metadata</code> elements
     */
    List<Element> getRecords(int startPosition, GetRecordsStatus status, Map<String, Set<String>> filters) throws IOException, DocumentException {

        final StringBuilder xmlRequest = new StringBuilder("<csw:GetRecords xmlns:csw=\"http://www.opengis.net/cat/csw/2.0.2\" " +
                "xmlns:ogc=\"http://www.opengis.net/ogc\" xmlns:apiso=\"http://www.opengis.net/cat/apiso/1.0\" ");
        xmlRequest.append(" startPosition=\"").append(startPosition).append("\" maxRecords=\"").append(RECORDS_PER_PAGE)
                .append("\" outputFormat=\"application/xml\"\n")
                .append(" resultType=\"results\" service=\"CSW\" version=\"2.0.2\" outputSchema=\"http://www.isotc211.org/2005/gmd\">");

        if (distributedSearch) {
            xmlRequest.append("<csw:DistributedSearch hopCount=\"1\"  />");
        }

        final String filterQuery = buildFilterQuery(filters);

        xmlRequest.append("<csw:Query typeNames=\"gmd:MD_Metadata\" xmlns:gmd=\"http://www.isotc211.org/2005/gmd\">")
                .append("<csw:ElementSetName>full</csw:ElementSetName>");

        if (StringUtils.isNotEmpty(filterQuery)) {
            xmlRequest.append("<csw:Constraint version=\"1.1.0\">")
                    .append("<ogc:Filter>\n")
                    .append(filterQuery)
                    .append("</ogc:Filter>\n")
                    .append("</csw:Constraint>\n");
        }

        if (sortResults) {
            xmlRequest.append("<ogc:SortBy>\n" +
                    "<ogc:SortProperty>\n" +
                    "<ogc:PropertyName>apiso:Identifier</ogc:PropertyName>\n" +
                    "<ogc:SortOrder>ASC</ogc:SortOrder>\n" +
                    "</ogc:SortProperty>\n" +
                    "</ogc:SortBy>");
        }

        xmlRequest.append("</csw:Query></csw:GetRecords>");

        final HttpPost request = new HttpPost(url);
        request.setHeader(HttpHeaders.CONTENT_TYPE, "application/xml");
        request.setHeader("X-CSW2DCATBridge", "getRecords");
        if (StringUtils.isNotEmpty(authorization)) {
            request.setHeader(HttpHeaders.AUTHORIZATION, authorization);
        }
        request.setEntity(new StringEntity(xmlRequest.toString(), StandardCharsets.UTF_8));

      /*  { FileWriter fw = new FileWriter("/tmp/request.xml");
        fw.write(xmlRequest.toString());
        fw.close();} */

        final long start = System.currentTimeMillis();
        try (CloseableHttpResponse response = httpClient.execute(request)) {
            if (response.getCode() == 200) {

                final Document document = saxReader.read(response.getEntity().getContent());

                log.debug("CSW request took {} ms", System.currentTimeMillis() - start);

                // dump response to a file for debugging
                if (dumpCswResponse != null) {
                    FileWriter fw = new FileWriter(dumpCswResponse);
                    document.write(fw);
                    fw.close();
                }

                if (verboseLogging != null) {
                    // Write the response into  file
                    final File dir = new File(verboseLogging);

                    if (dir.isDirectory()) {
                        final File file = new File(dir, "in_" + DateTimeFormatter.ISO_LOCAL_DATE_TIME.format(LocalDateTime.now()) + ".xml");
                        FileWriter fw = new FileWriter(file);
                        document.write(fw);
                        fw.close();
                    } else {
                        log.warn("Location for verbose logging {} is not a directory. No file written.", verboseLogging);
                    }
                }

                final Element searchResults = document.getRootElement().element("SearchResults");

                if (searchResults != null) {
                    if (status != null) {
                        status.nextRecord = NumberUtils.toInt(searchResults.attributeValue("nextRecord"));
                        status.numberOfRecordsMatched = NumberUtils.toInt(searchResults.attributeValue("numberOfRecordsMatched"));
                        status.numberOfRecordsReturned = NumberUtils.toInt(searchResults.attributeValue("numberOfRecordsReturned"));
                    }
                    return searchResults.elements();
                } else {
                    return Collections.emptyList();
                }
            } else {
                log.error("Response from CSW server was {} {}", response.getCode(), response.getReasonPhrase());
                return Collections.emptyList();
            }
        }
    }

    /**
     * Construct the content of the &lt;ogc:Filter> element.
     */
    String buildFilterQuery(Map<String, Set<String>> filters) {
        StringBuilder xmlRequest = new StringBuilder();

        if (filterOpendata) {
            if (filters.containsKey("Subject")) {
                filters.get("Subject").add("opendata");
            } else {
                filters.put("Subject", Collections.singleton("opendata"));
            }
        }

        // Do we have multiple CQL filter queries and therefore need a surrounding And element?
        final boolean needsAndElement = !filters.isEmpty() || hasMultipleTypeExclude() || hasTypeInclude();

        if (needsAndElement) {
            xmlRequest.append("<ogc:And>");
            for (String propertyName : filters.keySet()) {
                Set<String> values = filters.get(propertyName);
                for (String value : values) {
                    xmlRequest.append("<ogc:PropertyIsEqualTo><ogc:PropertyName>apiso:")
                            .append(propertyName)
                            .append("</ogc:PropertyName>")
                            .append("<ogc:Literal>")
                            .append(value)
                            .append("</ogc:Literal>")
                            .append("</ogc:PropertyIsEqualTo>");
                }
            }
        }

        if (hasTypeInclude()) {
            xmlRequest.append("<ogc:Or>");

            for (String type : typeInclude) {
                xmlRequest.append("<ogc:PropertyIsEqualTo>"
                                + "<ogc:PropertyName>apiso:Type</ogc:PropertyName>"
                                + "<ogc:Literal>")
                        .append(type)
                        .append("</ogc:Literal></ogc:PropertyIsEqualTo>");
            }

            xmlRequest.append("</ogc:Or>");
        }

        if (hasTypeExclude()) {
            for (String type : typeExclude) {
                xmlRequest.append("<ogc:Not><ogc:PropertyIsEqualTo>"
                                + "<ogc:PropertyName>apiso:Type</ogc:PropertyName>"
                                + "<ogc:Literal>")
                        .append(type)
                        .append("</ogc:Literal></ogc:PropertyIsEqualTo></ogc:Not>");
            }
        }


        if (needsAndElement) {
            xmlRequest.append("</ogc:And>");
        }

        return xmlRequest.toString();
    }

    List<Element> findOperatesOn(String id) throws IOException, DocumentException {
        final StringBuilder filter = new StringBuilder();

        if (filterOpendata) {
            filter.append("<ogc:And>");
        }
        filter.append("        <ogc:PropertyIsEqualTo>" +
                        "          <ogc:PropertyName>operatesOn</ogc:PropertyName>" +
                        "          <ogc:Literal>")
                .append(StringEscapeUtils.escapeXml11(id))
                .append("</ogc:Literal>")
                .append("        </ogc:PropertyIsEqualTo>\n");

        if (filterOpendata) {
            filter.append("<ogc:PropertyIsEqualTo>" +
                    "<ogc:PropertyName xmlns:apiso=\"http://www.opengis.net/cat/apiso/1.0\">apiso:Subject</ogc:PropertyName>" +
                    "<ogc:Literal>opendata</ogc:Literal>" +
                    "</ogc:PropertyIsEqualTo>");
            filter.append("</ogc:And>");
        }

        final String xmlRequest = "<csw:GetRecords xmlns:csw=\"http://www.opengis.net/cat/csw/2.0.2\" xmlns:ogc=\"http://www.opengis.net/ogc\"\n" +
                "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:ows=\"http://www.opengis.net/ows\"\n" +
                "xmlns:apiso=\"http://www.opengis.net/cat/apiso/1.0\" " +
                "outputSchema=\"http://www.isotc211.org/2005/gmd\" outputFormat=\"application/xml\" version=\"2.0.2\" service=\"CSW\" resultType=\"results\"\n" +
                "maxRecords=\"999\" xsi:schemaLocation=\"http://www.opengis.net/cat/csw/2.0.2 http://schemas.opengis.net/csw/2.0.2/CSW-discovery.xsd\">\n" +
                "  <csw:Query typeNames=\"csw:Record\">\n" +
                "    <csw:ElementSetName>full</csw:ElementSetName>\n" +
                "    <csw:Constraint version=\"1.1.0\">\n" +
                "      <ogc:Filter>\n" +
                filter +
                "      </ogc:Filter>\n" +
                "    </csw:Constraint>\n" +
                "  </csw:Query>\n" +
                "</csw:GetRecords>";

        final HttpPost request = new HttpPost(url);
        request.setHeader(HttpHeaders.CONTENT_TYPE, "text/xml");
        request.setHeader("X-CSW2DCATBridge", "operatesOn");
        if (StringUtils.isNotEmpty(authorization)) {
            request.setHeader(HttpHeaders.AUTHORIZATION, authorization);
        }
        request.setEntity(new StringEntity(xmlRequest, StandardCharsets.UTF_8));

        try (CloseableHttpResponse response = httpClient.execute(request)) {
            final Document document = saxReader.read(response.getEntity().getContent());

            if (document == null || document.getRootElement().elements("SearchResults").isEmpty()) {
                return Collections.emptyList();
            }

            return document.getRootElement().element("SearchResults").elements();
        }
    }


    /**
     * Holds information about the results of a getRecords operation.
     */
    public static class GetRecordsStatus {
        private int nextRecord;
        private int numberOfRecordsMatched;
        private int numberOfRecordsReturned;

        public int getNextRecord() {
            return nextRecord;
        }

        public int getNumberOfRecordsMatched() {
            return numberOfRecordsMatched;
        }

        public int getNumberOfRecordsReturned() {
            return numberOfRecordsReturned;
        }
    }
}
