package de.landsh.opendata.csw2dcat;

import de.landsh.opendata.csw2dcat.vocabulary.Locn;
import de.landsh.opendata.csw2dcat.vocabulary.SchemaOrg;
import org.apache.commons.lang3.StringUtils;
import org.apache.jena.rdf.model.*;
import org.apache.jena.vocabulary.DCAT;
import org.apache.jena.vocabulary.DCTerms;
import org.apache.jena.vocabulary.VCARD4;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.QName;
import org.json.JSONArray;
import org.json.JSONObject;
import org.json.JSONTokener;
import org.locationtech.jts.geom.Coordinate;
import org.locationtech.jts.geom.Geometry;
import org.locationtech.jts.io.ParseException;
import org.locationtech.jts.io.WKTReader;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

public class Dataset2MDMetadata {
    public static final Namespace NS_GMD = Namespace.get("gmd", "http://www.isotc211.org/2005/gmd");
    public static final Namespace NS_GML = Namespace.get("gml", "http://www.opengis.net/gml/3.2");
    public static final Namespace NS_GCO = Namespace.get("gco", "http://www.isotc211.org/2005/gco");
    public static final Namespace NS_GMX = Namespace.get("gmx", "http://www.isotc211.org/2005/gmx");
    public static final Namespace NS_XLINK = Namespace.get("xlink", "http://www.w3.org/1999/xlink");
    private static final Logger log = LoggerFactory.getLogger(Dataset2MDMetadata.class);

    private final Map<String, JSONObject> url2license = new HashMap<>();

    public Dataset2MDMetadata() {
        JSONArray licenses = new JSONArray(new JSONTokener(Objects.requireNonNull(getClass().getResourceAsStream("/licenses.json"))));
        for (Object l : licenses) {
            JSONObject license = (JSONObject) l;
            url2license.put(license.getString("uri"), license);
        }
    }

    /**
     * Converts a URI describing a DCAT-AP language into an ISO 639-2 language code.
     */
    static String mapLanguageCode(String uri) {
        if ("http://publications.europa.eu/resource/authority/language/DEU".equals(uri)) {
            return "ger";
        } else {
            return StringUtils.lowerCase(StringUtils.substringAfterLast(uri, "/"));
        }
    }

    static Element gcoCharacterString(String value) {
        Element element = DocumentHelper.createElement(QName.get("CharacterString", NS_GCO));
        element.addText(value);
        return element;
    }

    private static String getValueOrNull(Resource resource, Property property) {
        if (resource.hasProperty(property)) {
            Statement stmt = resource.getProperty(property);
            if (stmt.getObject().isLiteral()) {
                return stmt.getLiteral().getString();
            } else {
                return stmt.getResource().getURI();
            }
        }
        return null;
    }

    private static void addLanguageIfPresent(Resource dataset, Element element) {
        final Resource language = dataset.getPropertyResourceValue(DCTerms.language);
        if (language != null) {
            final String languageCode = mapLanguageCode(language.getURI());

            element.addElement(QName.get("language", NS_GMD))
                    .addElement(QName.get("LanguageCode", NS_GMD))
                    .addAttribute("codeList", "http://www.loc.gov/standards/iso639-2/")
                    .addAttribute("codeListValue", languageCode);
        }
    }

    /**
     * If the specified text is not empty add an element with the name specified in elementName and a child element
     * <code>gco:CharacterString</code> with the specified text to targetElement.
     */
    private static void addElementWithCharacterStringIfPresent(Element targetElement, QName elementName, String text) {
        if (StringUtils.isNotEmpty(text)) {
            targetElement.addElement(elementName)
                    .addElement(QName.get("CharacterString", NS_GCO))
                    .addText(text);
        }
    }

    String licenseAsText(String licenseURI, String licenseAttributionByText) {
        final JSONObject license = url2license.get(licenseURI);
        if (license != null) {
            if (license.has("text")) {
                return license.getString("text") + licenseAttributionByText +".";
            } else {
                StringBuilder result = new StringBuilder("Für die Nutzung der Daten ist die ")
                        .append(license.getString("name"))
                        .append(" anzuwenden. Die Lizenz ist über ")
                        .append(license.getString("url"))
                        .append(" anzuwenden.");

                if (StringUtils.isNotBlank(licenseAttributionByText)) {
                    result.append(" Der Quellenvermerk lautet: ")
                            .append(licenseAttributionByText)
                            .append(".");
                }

                return result.toString();
            }
        } else {
            throw new IllegalArgumentException("Unknown license " + licenseURI);
        }
    }

    String licenseAsJSON(String licenseURI, String licenseAttributionByText) {

        final JSONObject license = url2license.get(licenseURI);
        if (license != null) {
            StringBuilder result = new StringBuilder("{\"id\":\"")
                    .append(license.getString("id"))
                    .append("\",\"name\":\"")
                    .append(license.getString("name"))
                    .append("\",\"url\":\"")
                    .append(license.getString("url"))
                    .append("\"");

            if (StringUtils.isNotBlank(licenseAttributionByText)) {
                result.append(",\"quelle\":\"")
                        .append(licenseAttributionByText)
                        .append("\"");
            }

            result.append("}");
            return result.toString();

        } else {
            throw new IllegalArgumentException("Unknown license " + licenseURI);
        }
    }

    /**
     * Convert a DCAT-AP.de compliant dcat:Dataset {@link Resource} into an {@link Element} containing a
     * gmd:MD_Metadata element
     *
     * @param dataset the    DCAT-AP.de compliant dcat:Dataset
     * @return an {@link Element} containing a gmd:MD_Metadata element
     */
    public Element convert(Resource dataset) {
        final Element element = DocumentHelper.createElement(QName.get("MD_Metadata", NS_GMD));
        element.add(NS_GCO);
        element.add(NS_GML);
        element.add(NS_GMX);

        final String identifier = dataset.getProperty(DCTerms.identifier).getString();
        element.addElement(QName.get("fileIdentifier", NS_GMD)).add(gcoCharacterString(identifier));

        addLanguageIfPresent(dataset, element);

        element.addElement(QName.get("characterSet", NS_GMD))
                .addElement(QName.get("MD_CharacterSetCode", NS_GMD))
                .addAttribute("codeList", "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#MD_CharacterSetCode")
                .addAttribute("codeListValue", "utf8");

        element.addElement(QName.get("hierarchyLevel", NS_GMD))
                .addElement(QName.get("MD_ScopeCode", NS_GMD))
                .addAttribute("codeList", "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#MD_ScopeCode")
                .addAttribute("codeListValue", "dataset");


        final Element contact = element.addElement(QName.get("contact", NS_GMD));
        final Resource contactPoint = dataset.getPropertyResourceValue(DCAT.contactPoint);
        if (contactPoint == null) {
            contact.addAttribute(QName.get("nilReason", NS_GCO), "unknown");
        } else {
            contact.add(contactPoint(contactPoint));
        }

        element.addElement(QName.get("dateStamp", NS_GMD))
                .addElement(QName.get("Date", NS_GCO))
                .addText(dataset.getProperty(DCTerms.modified).getString().substring(0, 10));

        element.addElement(QName.get("metadataStandardName", NS_GMD))
                .addElement(QName.get("CharacterString", NS_GCO))
                .addText("ISO19115");

        element.addElement(QName.get("metadataStandardVersion", NS_GMD))
                .addElement(QName.get("CharacterString", NS_GCO))
                .addText("2003/Cor.1:2006");

        referenceSystem(dataset, element);

        identificationInfo(dataset, element);

        distributionInfo(dataset, element);

        dataQualityInfo(dataset, element);

        return element;
    }

    private void referenceSystem(Resource dataset, Element targetElement) {
        StmtIterator it = dataset.listProperties(DCTerms.conformsTo);
        while (it.hasNext()) {
            String uri = it.next().getResource().getURI();
            if (uri.startsWith("http://www.opengis.net/def/crs/EPSG/0/")) {
                String code = StringUtils.substringAfterLast(uri, "/");
                targetElement.addElement(QName.get("referenceSystemInfo", NS_GMD))
                        .addElement(QName.get("MD_ReferenceSystem", NS_GMD))
                        .addElement(QName.get("referenceSystemIdentifier", NS_GMD))
                        .addElement(QName.get("RS_Identifier", NS_GMD))
                        .addElement(QName.get("code", NS_GMD))
                        .addElement(QName.get("Anchor", NS_GMX))
                        .addAttribute(QName.get("href", NS_XLINK), uri)
                        .addText(code);
            }
        }

    }

    private Element contactPoint(Resource contactPoint) {
        Element ciResponsibleParty = DocumentHelper.createElement(QName.get("CI_ResponsibleParty", NS_GMD));

        String fn = getValueOrNull(contactPoint, VCARD4.fn);
        String email = StringUtils.replace(getValueOrNull(contactPoint, VCARD4.hasEmail), "mailto:", "");
        String url = getValueOrNull(contactPoint, VCARD4.hasURL);
        String telephone = getValueOrNull(contactPoint, VCARD4.hasTelephone);
        String streetAddress = getValueOrNull(contactPoint, VCARD4.hasStreetAddress);
        String city = getValueOrNull(contactPoint, VCARD4.hasLocality);
        String postalCode = getValueOrNull(contactPoint, VCARD4.hasPostalCode);
        String countryName = getValueOrNull(contactPoint, VCARD4.hasCountryName);

        ciResponsibleParty.addElement(QName.get("organisationName", NS_GMD))
                .addElement(QName.get("CharacterString", NS_GCO))
                .addText(fn);

        Element ciContact = ciResponsibleParty.addElement(QName.get("contactInfo", NS_GMD))
                .addElement(QName.get("CI_Contact", NS_GMD));

        if (StringUtils.isNotEmpty(telephone)) {
            ciContact.addElement(QName.get("phone", NS_GMD))
                    .addElement(QName.get("CI_Telephone", NS_GMD))
                    .addElement(QName.get("voice", NS_GMD))
                    .addElement(QName.get("CharacterString", NS_GCO))
                    .addText(telephone);
        }

        Element ciAddress = ciContact.addElement(QName.get("address", NS_GMD))
                .addElement(QName.get("CI_Address", NS_GMD));
        addElementWithCharacterStringIfPresent(ciAddress, QName.get("deliveryPoint", NS_GMD), streetAddress);
        addElementWithCharacterStringIfPresent(ciAddress, QName.get("city", NS_GMD), city);
        addElementWithCharacterStringIfPresent(ciAddress, QName.get("postalCode", NS_GMD), postalCode);
        addElementWithCharacterStringIfPresent(ciAddress, QName.get("country", NS_GMD), countryName);
        addElementWithCharacterStringIfPresent(ciAddress, QName.get("electronicMailAddress", NS_GMD), email);

        if (StringUtils.isNotEmpty(url)) {
            ciContact.addElement(QName.get("onlineResource", NS_GMD))
                    .addElement(QName.get("CI_OnlineResource", NS_GMD))
                    .addElement(QName.get("linkage", NS_GMD))
                    .addElement(QName.get("URL", NS_GMD))
                    .addText(url);
        }

        //TODO <xsl:apply-templates select="vcard:Individual/vcard:fn" mode="individual"/>
        //TODO <xsl:call-template name="contactInfo"/>
        ciResponsibleParty.addElement(QName.get("role", NS_GMD))
                .addElement(QName.get("CI_RoleCode", NS_GMD))
                .addAttribute("codeList", "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#CI_RoleCode")
                .addAttribute("codeListValue", "pointOfContact")
                .addText("pointOfContact");
        return ciResponsibleParty;
    }

    public void identificationInfo(Resource dataset, Element targetElement) {
        Element mdDataIdentification = targetElement.addElement(QName.get("identificationInfo", NS_GMD))
                .addElement(QName.get("MD_DataIdentification", NS_GMD));

        Element citation = mdDataIdentification.addElement(QName.get("citation", NS_GMD));

        Element ciCitation = citation.addElement(QName.get("CI_Citation", NS_GMD));

        // === dct:title ======================
        ciCitation.addElement(QName.get("title", NS_GMD))
                .addElement(QName.get("CharacterString", NS_GCO))
                .addText(dataset.getProperty(DCTerms.title).getString());

        // === dct:issued ======================
        String date = getValueOrNull(dataset, DCTerms.issued).substring(0, 10);
        Element ciDate = ciCitation.addElement(QName.get("date", NS_GMD))
                .addElement(QName.get("CI_Date", NS_GMD));
        ciDate.addElement(QName.get("date", NS_GMD))
                .addElement(QName.get("Date", NS_GCO)).addText(date);
        ciDate.addElement(QName.get("dateType", NS_GMD))
                .addElement(QName.get("CI_DateTypeCode", NS_GMD))
                .addAttribute("codeList", "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#CI_DateTypeCode")
                .addAttribute("codeListValue", "publication");

        // === dct:identifier ======================
        String id = getValueOrNull(dataset, DCTerms.identifier);
        ciCitation.addElement(QName.get("identifier", NS_GMD))
                .addElement(QName.get("MD_Identifier", NS_GMD))
                .addElement(QName.get("code", NS_GMD))
                .addElement(QName.get("CharacterString", NS_GCO))
                .addText(id);

        // TODO dct:created
        // TODO dct:modified

        if (dataset.hasProperty(DCTerms.description)) {
            mdDataIdentification.addElement(QName.get("abstract", NS_GMD))
                    .addElement(QName.get("CharacterString", NS_GCO))
                    .addText(dataset.getProperty(DCTerms.description).getString());
        }

        final Element pointOfContact = mdDataIdentification.addElement(QName.get("pointOfContact", NS_GMD));
        final Resource contactPoint = dataset.getPropertyResourceValue(DCAT.contactPoint);
        if (contactPoint == null) {
            pointOfContact.addAttribute(QName.get("nilReason", NS_GCO), "unknown");
        } else {
            pointOfContact.add(contactPoint(contactPoint));
        }

        // TODO <xsl:apply-templates select="dct:publisher"/>
        // TODO <xsl:apply-templates select="dcatde:maintainer"/>
        // TODO <xsl:apply-templates select="dct:creator"/>
        // TODO <xsl:apply-templates select="dct:rightsHolder"/>

        Element mdKeywords = mdDataIdentification.addElement(QName.get("descriptiveKeywords", NS_GMD))
                .addElement(QName.get("MD_Keywords", NS_GMD));

        StmtIterator itKeyword = dataset.listProperties(DCAT.keyword);
        while (itKeyword.hasNext()) {
            mdKeywords.addElement(QName.get("keyword", NS_GMD))
                    .addElement(QName.get("CharacterString", NS_GCO))
                    .addText(itKeyword.next().getString());
        }

        // === <gmd:accessConstraints> ============================
        {
            Element mdLegalConstraints = mdDataIdentification.addElement(QName.get("resourceConstraints", NS_GMD))
                    .addElement(QName.get("MD_LegalConstraints", NS_GMD));
            mdLegalConstraints.addElement(QName.get("accessConstraints", NS_GMD))
                    .addElement(QName.get("MD_RestrictionCode", NS_GMD))
                    .addAttribute("codeList", "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_RestrictionCode")
                    .addAttribute("codeListValue", "otherRestrictions");
            mdLegalConstraints.addElement(QName.get("otherConstraints", NS_GMD))
                    .addElement(QName.get("Anchor", NS_GMX))
                    .addAttribute(QName.get("href", NS_XLINK), "http://inspire.ec.europa.eu/metadata-codelist/LimitationsOnPublicAccess/noLimitations")
                    .addText("Es gelten keine Zugriffsbeschränkungen");
        }

        // === <gmd:useConstraints> ============================
        {
            final String licenseURI = getValueOrNull(dataset, DCTerms.license);
            final String licenseAttributionByText = getValueOrNull(dataset, ResourceFactory.createProperty("http://dcat-ap.de/def/dcatde/", "licenseAttributionByText"));
            Element mdLegalConstraints = mdDataIdentification.addElement(QName.get("resourceConstraints", NS_GMD))
                    .addElement(QName.get("MD_LegalConstraints", NS_GMD));
            mdLegalConstraints.addElement(QName.get("useConstraints", NS_GMD))
                    .addElement(QName.get("MD_RestrictionCode", NS_GMD))
                    .addAttribute("codeList", "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_RestrictionCode")
                    .addAttribute("codeListValue", "otherRestrictions");
            mdLegalConstraints.addElement(QName.get("otherConstraints", NS_GMD))
                    .addElement(QName.get("CharacterString", NS_GCO))
                    .addText(licenseAsText(licenseURI, licenseAttributionByText));
            mdLegalConstraints.addElement(QName.get("otherConstraints", NS_GMD))
                    .addElement(QName.get("CharacterString", NS_GCO))
                    .addText(licenseAsJSON(licenseURI, licenseAttributionByText));
        }

        addLanguageIfPresent(dataset, mdDataIdentification);

        // === dct:temporal ============================
        if (dataset.hasProperty(DCTerms.temporal)) {
            Resource periodOfTime = dataset.getProperty(DCTerms.temporal).getResource();
            Element timePeriod = mdDataIdentification.addElement(QName.get("extent", NS_GMD))
                    .addElement(QName.get("EX_Extent", NS_GMD))
                    .addElement(QName.get("temporalElement", NS_GMD))
                    .addElement(QName.get("EX_TemporalExtent", NS_GMD))
                    .addElement(QName.get("extent", NS_GMD))
                    .addElement(QName.get("TimePeriod", NS_GML));

            if (periodOfTime.hasProperty(DCAT.startDate)) {
                String value = StringUtils.substring(periodOfTime.getProperty(DCAT.startDate).getLiteral().getString(), 0, 10);
                timePeriod.addElement(QName.get("beginPosition", NS_GML))
                        .addText(value);
            } else if (periodOfTime.hasProperty(SchemaOrg.startDate)) {
                String value = StringUtils.substring(periodOfTime.getProperty(SchemaOrg.startDate).getLiteral().getString(), 0, 10);
                timePeriod.addElement(QName.get("beginPosition", NS_GML))
                        .addText(value);
            } else {
                timePeriod.addElement(QName.get("beginPosition", NS_GML))
                        .addAttribute("indeterminatePosition", "unknown");
            }

            if (periodOfTime.hasProperty(DCAT.endDate)) {
                String value = StringUtils.substring(periodOfTime.getProperty(DCAT.endDate).getLiteral().getString(), 0, 10);
                timePeriod.addElement(QName.get("endPosition", NS_GML))
                        .addText(value);
            } else if (periodOfTime.hasProperty(SchemaOrg.endDate)) {
                String value = StringUtils.substring(periodOfTime.getProperty(SchemaOrg.endDate).getLiteral().getString(), 0, 10);
                timePeriod.addElement(QName.get("endPosition", NS_GML))
                        .addText(value);
            } else {
                timePeriod.addElement(QName.get("endPosition", NS_GML))
                        .addAttribute("indeterminatePosition", "unknown");
            }
        }

        // === dct:spatial ============================
        if (dataset.hasProperty(DCTerms.spatial)) {
            Resource location = dataset.getProperty(DCTerms.spatial).getResource();
            StmtIterator it = location.listProperties(Locn.geometry);
            while (it.hasNext()) {
                Literal geometryLiteral = it.next().getLiteral();
                if ("http://www.opengis.net/ont/geosparql#wktLiteral".equals(geometryLiteral.getDatatypeURI())) {
                    String wkt = geometryLiteral.getString();

                    try {
                        Geometry geom = new WKTReader().read(wkt);
                        Coordinate[] envelope = geom.getEnvelope().getCoordinates();

                        double north = envelope[2].y;
                        double south = envelope[0].y;
                        double west = envelope[0].x;
                        double east = envelope[2].x;

                        Element EX_GeographicBoundingBox = mdDataIdentification
                                .addElement(QName.get("extent", NS_GMD))
                                .addElement(QName.get("EX_Extent", NS_GMD))
                                .addElement(QName.get("geographicElement", NS_GMD))
                                .addElement(QName.get("EX_GeographicBoundingBox", NS_GMD));
                        EX_GeographicBoundingBox
                                .addElement(QName.get("extentTypeCode", NS_GMD))
                                .addElement(QName.get("Boolean", NS_GCO))
                                .addText("true");
                        EX_GeographicBoundingBox
                                .addElement(QName.get("westBoundLongitude", NS_GMD))
                                .addElement(QName.get("Decimal", NS_GCO))
                                .addText(Double.toString(west));
                        EX_GeographicBoundingBox
                                .addElement(QName.get("eastBoundLongitude", NS_GMD))
                                .addElement(QName.get("Decimal", NS_GCO))
                                .addText(Double.toString(east));
                        EX_GeographicBoundingBox
                                .addElement(QName.get("southBoundLatitude", NS_GMD))
                                .addElement(QName.get("Decimal", NS_GCO))
                                .addText(Double.toString(south));
                        EX_GeographicBoundingBox
                                .addElement(QName.get("northBoundLatitude", NS_GMD))
                                .addElement(QName.get("Decimal", NS_GCO))
                                .addText(Double.toString(north));
                    } catch (ParseException e) {
                        log.warn("Could not parse WKT for dataset {}.", id);
                    }
                }
            }
        }
    }

    String getDistributionFormat(Resource distribution) {
        final Statement formatStatement = distribution.getProperty(DCTerms.format);
        if( formatStatement != null && formatStatement.getObject().isLiteral()) {
            return formatStatement.getLiteral().getString();
        } else if (formatStatement != null ) {
            final String formatURI = formatStatement.getResource().getURI();
            return StringUtils.substringAfterLast(formatURI, "/");
        }

        return null;
    }

    public void distributionInfo(Resource dataset, Element targetElement) {

        Element mdDistribution = targetElement.addElement(QName.get("distributionInfo", NS_GMD))
                .addElement(QName.get("MD_Distribution", NS_GMD));

        Set<String> formats = new HashSet<>();
        StmtIterator it = dataset.listProperties(DCAT.distribution);
        while (it.hasNext()) {
            final Resource distribution = it.next().getResource();
            final String format = getDistributionFormat(distribution);
            if( format != null) {
                formats.add(format);
            }
        }
        for (String format : formats) {
            Element mdFormat =
                    mdDistribution.addElement(QName.get("distributionFormat", NS_GMD))
                            .addElement(QName.get("MD_Format", NS_GMD));
            mdFormat
                    .addElement(QName.get("name", NS_GMD))
                    .addElement(QName.get("CharacterString", NS_GCO))
                    .addText(format);
            mdFormat.addElement(QName.get("version", NS_GMD))
                    .addAttribute(QName.get("nilReason", NS_GCO), "unknown");
        }

        it = dataset.listProperties(DCAT.distribution);
        while (it.hasNext()) {
            final Resource distribution = it.next().getResource();
            String format =getDistributionFormat(distribution);
            String name = getValueOrNull(distribution, DCTerms.title);
            String description = getValueOrNull(distribution, DCTerms.description);
            String downloadURL = getValueOrNull(distribution, DCAT.downloadURL);
            String accessURL = getValueOrNull(distribution, DCAT.accessURL);
            if(format != null) {
                formats.add(format);
            }

            Element ciOnlineResource = mdDistribution.addElement(QName.get("transferOptions", NS_GMD))
                    .addElement(QName.get("MD_DigitalTransferOptions", NS_GMD))
                    .addElement(QName.get("onLine", NS_GMD))
                    .addElement(QName.get("CI_OnlineResource", NS_GMD));

            if (StringUtils.isNotEmpty(downloadURL)) {
                ciOnlineResource.addElement(QName.get("linkage", NS_GMD))
                        .addElement(QName.get("URL", NS_GMD))
                        .addText(downloadURL);
            } else {
                if (StringUtils.isNotEmpty(accessURL)) {
                    ciOnlineResource.addElement(QName.get("linkage", NS_GMD))
                            .addElement(QName.get("URL", NS_GMD))
                            .addText(accessURL);
                }
            }
            if( format != null) {
                ciOnlineResource.addElement(QName.get("applicationProfile", NS_GMD))
                        .addElement(QName.get("CharacterString", NS_GCO))
                        .addText("http://publications.europa.eu/resource/authority/file-type/" + format.toUpperCase());
            }
            if( StringUtils.isNotBlank(name)) {
                ciOnlineResource.addElement(QName.get("name", NS_GMD))
                        .addElement(QName.get("CharacterString", NS_GCO))
                        .addText(name);
            }
            ciOnlineResource.addElement(QName.get("function", NS_GMD))
                    .addElement(QName.get("CI_OnLineFunctionCode", NS_GMD))
                    .addAttribute("codeList", "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#CI_OnLineFunctionCode")
                    .addAttribute("codeListValue", "download")
                    .addText("download");
        }


    }

    public void dataQualityInfo(Resource dataset, Element targetElement) {
        Element dataQuality = targetElement.addElement(QName.get("dataQualityInfo", NS_GMD))
                .addElement(QName.get("DQ_DataQuality", NS_GMD));

        dataQuality.addElement(QName.get("scope", NS_GMD))
                .addElement(QName.get("DQ_Scope", NS_GMD))
                .addElement(QName.get("level", NS_GMD))
                .addElement(QName.get("MD_ScopeCode", NS_GMD))
                .addAttribute("codeList", "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#MD_ScopeCode")
                .addAttribute("codeListValue", "dataset");

        if (dataset.hasProperty(DCTerms.provenance)) {
            dataQuality.addElement(QName.get("lineage", NS_GMD))
                    .addElement(QName.get("LI_Lineage", NS_GMD));
            // TODO <xsl:call-template name="gcoCharacterStringResource"/>
        }

        if (dataset.hasProperty(DCTerms.conformsTo)) {
            //TODO
        }

    }
}
