package de.landsh.opendata.csw2dcat;

import io.undertow.server.HttpServerExchange;
import io.undertow.util.Headers;
import io.undertow.util.HttpString;
import org.apache.commons.lang3.StringUtils;
import org.apache.hc.client5.http.classic.methods.HttpGet;
import org.apache.hc.client5.http.impl.classic.CloseableHttpClient;
import org.apache.hc.client5.http.impl.classic.CloseableHttpResponse;
import org.apache.hc.client5.http.impl.classic.HttpClients;
import org.apache.hc.core5.http.HttpStatus;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.ResIterator;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.vocabulary.DCAT;
import org.apache.jena.vocabulary.RDF;
import org.dom4j.Element;
import org.json.JSONArray;
import org.json.JSONObject;
import org.json.JSONTokener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URL;
import java.util.SortedSet;
import java.util.TreeSet;

public class Dcat2iso {
    private static final Logger log = LoggerFactory.getLogger(Dcat2iso.class);
    private final String dcatDatasetBaseURL;
    private final String mode;
    private final String ckanApiURL;
    private final String baseURL;
    private CloseableHttpClient httpClient = HttpClients.createMinimal();

    public Dcat2iso(String mode, String baseURL, String dcatDatasetBaseURL, String ckanApiURL) {
        this.dcatDatasetBaseURL = dcatDatasetBaseURL;
        this.mode = mode;
        this.baseURL = baseURL;
        this.ckanApiURL = ckanApiURL;

        if ("ckan".equals(mode) && StringUtils.isEmpty(ckanApiURL)) {
            throw new IllegalArgumentException("CKAN API URL must be set if the mode for the DCAT->ISO is 'ckan'.");
        }
    }

    /**
     * Handle DCAT -> ISO request
     */
    void handleDcat2isoRequest(HttpServerExchange exchange) {
        String id = StringUtils.substringAfter(exchange.getRequestPath(), "/iso/");
        try {
            URL url = new URL(dcatDatasetBaseURL + id + ".rdf");

            final Model model = ModelFactory.createDefaultModel();
            try (InputStream in = url.openStream()) {
                model.read(in, null);
            }

            ResIterator it = model.listResourcesWithProperty(RDF.type, DCAT.Dataset);
            if (it.hasNext()) {
                final Resource dataset = it.next();
                final Element isoMetadata = new Dataset2MDMetadata().convert(dataset);

                final StringWriter out = new StringWriter();
                isoMetadata.write(out);
                out.close();

                exchange.setStatusCode(HttpStatus.SC_OK);
                exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "application/xml");
                exchange.getResponseHeaders().put(HttpString.tryFromString("Access-Control-Allow-Origin"), "*");
                exchange.getResponseSender().send(out.toString());
            } else {
                exchange.setStatusCode(HttpStatus.SC_NOT_FOUND);
                exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "text/plain");
                exchange.getResponseSender().send("Dataset with id " + id + " not Found");
            }

        } catch (IOException e) {
            exchange.setStatusCode(HttpStatus.SC_INTERNAL_SERVER_ERROR);
            exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "text/plain");
            exchange.getResponseSender().send("Internal server error. \n" + e.getMessage());
            log.warn("Could not convert DCAT metadata to ISO: {}", e.getMessage());
        }
    }

    /**
     * Create a list of URI with datasets marked for ISO metadata export (e.g. tag "GDI-SH")
     */
    void list(HttpServerExchange exchange, String queryTag) {
        SortedSet<String> ids = new TreeSet<>();
        if ("ckan".equals(mode)) {
            // use the CKAN API to retrieve matching datasets
            HttpGet get = new HttpGet(ckanApiURL + "/action/package_search?rows=1000&fq=tags:" + queryTag);
            try (CloseableHttpResponse response = httpClient.execute(get)) {
                JSONObject json = new JSONObject(new JSONTokener(response.getEntity().getContent()));
                JSONArray datasets = json.getJSONObject("result").getJSONArray("results");
                for (int i = 0; i < datasets.length(); i++) {
                    JSONObject dataset = datasets.getJSONObject(i);
                    ids.add(dataset.getString("id"));
                }
            } catch (IOException e) {
                exchange.setStatusCode(HttpStatus.SC_INTERNAL_SERVER_ERROR);
                exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "text/plain");
                exchange.getResponseSender().send("Internal server error. \n" + e.getMessage());
            }
        }

        JSONArray result = new JSONArray();
        for (String id : ids) {
            result.put(baseURL + "/iso/" + id);
        }
        exchange.setStatusCode(HttpStatus.SC_OK);
        exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "application/json");
        exchange.getResponseSender().send(result.toString());
    }

}
