package de.landsh.opendata.csw2dcat;


import org.apache.commons.lang3.StringUtils;
import org.apache.jena.datatypes.BaseDatatype;
import org.apache.jena.datatypes.xsd.XSDDatatype;
import org.apache.jena.rdf.model.*;
import org.apache.jena.vocabulary.RDF;
import org.dom4j.Element;
import org.dom4j.QName;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class GenericMetadataConverter {

    public static final String NAMESPACE_GCO = "http://www.isotc211.org/2005/gco";
    public static final String NAMESPACE_GMD = "http://www.isotc211.org/2005/gmd";
    public static final String NAMESPACE_GMX = "http://www.isotc211.org/2005/gmx";
    private final static char NAMESPACE_SEPERATOR = '#';
    private static final String NAMESPACE_XLINK = "http://www.w3.org/1999/xlink";
    Logger log = LoggerFactory.getLogger(GenericMetadataConverter.class);

    public Resource iso2dcat(Element metadata) {
        if ("GetRecordByIdResponse".equals(metadata.getName())) {
            metadata = metadata.element("MD_Metadata");
        }

        if (!"MD_Metadata".equals(metadata.getName())) {
            throw new IllegalArgumentException("Only a gmd:MD_Metadata element can be converted.");
        }

        Model model = ModelFactory.createDefaultModel();

        RDFNode node = workOnEntity(model, metadata);

        return (Resource) node;
    }

    boolean useHrefForAnchor = false;

    private RDFNode workOnEntity(Model model, Element element) {
        final String localName = element.getName();
        RDFNode result = null;
        if (element.elements().isEmpty()) {
            // a leaf node
            if (NAMESPACE_GCO.equals(element.getNamespaceURI())) {
                if ("CharacterString".equals(localName)) {
                    result = ResourceFactory.createStringLiteral(element.getText());
                } else if ("Boolean".equals(localName)) {
                    result = ResourceFactory.createTypedLiteral(element.getText(), XSDDatatype.XSDboolean);
                } else if ("Date".equals(localName)) {
                    result = ResourceFactory.createTypedLiteral(element.getText(), XSDDatatype.XSDdate);
                } else if ("Decimal".equals(localName)) {
                    result = ResourceFactory.createTypedLiteral(element.getText(), XSDDatatype.XSDdecimal);
                } else if ("LocalName".equals(localName)) {
                    result = ResourceFactory.createTypedLiteral(element.getText(), new BaseDatatype(element.attributeValue("codeSpace")));
                    //TODO is this mapping correct?
                } else {
                    log.warn("Unimplemented GCO element {}", localName);
                }
            } else if (NAMESPACE_GMD.equals(element.getNamespaceURI()) && "URL".equals(localName)) {
                result = ResourceFactory.createTypedLiteral(element.getText(), XSDDatatype.XSDanyURI);
            } else if (NAMESPACE_GMX.equals(element.getNamespaceURI()) && "Anchor".equals(localName)) {
                String href =   element.attributeValue(QName.get("href", NAMESPACE_XLINK));

                if( useHrefForAnchor && href != null) {
                    result = model.createResource( href);
                }   else {
                    result = ResourceFactory.createStringLiteral(element.getText());
                }
        }else {
                String codeList = element.attributeValue("codeList");
                String codeListValue = element.attributeValue("codeListValue");
                String uri = codeListValueToURI(codeList, codeListValue);

                Resource resource;
                if (uri != null) {
                    resource = model.createResource(uri);
                } else {
                    resource = model.createResource();
                }

                resource.addProperty(RDF.type, model.createResource(element.getNamespaceURI() + NAMESPACE_SEPERATOR + element.getName()));

                result = resource;
            }
        } else {
            Resource resource = model.createResource();
            resource.addProperty(RDF.type, model.createResource(element.getNamespaceURI() + NAMESPACE_SEPERATOR + element.getName()));
            result = resource;
            for (Element child : element.elements()) {
                workOnProperty(resource, child);
            }
        }
        return result;
    }

    String codeListValueToURI(String codeList, String codeListValue) {
        String uri = null;
        if (StringUtils.startsWith(codeList, "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#")
                || StringUtils.startsWith(codeList, "https://standards.iso.org/iso/19139/resources/gmxCodelists.xml#")) {
            // these code list entries are all built similarly
            uri = codeList + '_' + codeListValue;
        } else if ("http://www.loc.gov/standards/iso639-2/".equals(codeList)) {
            uri = codeList + codeListValue;
        } else {
            log.warn("Unknown code list {}", codeList);
        }
        return uri;
    }

    private void workOnProperty(Resource resource, Element element) {
        final Model model = resource.getModel();
        final int numberOfChildElements = element.elements().size();
        final Property property = model.createProperty(element.getNamespaceURI() + NAMESPACE_SEPERATOR + element.getName());

        if (numberOfChildElements == 0) {
            final String href = element.attributeValue(QName.get("href", NAMESPACE_XLINK));
            final String nilReason = element.attributeValue(QName.get("nilReason", NAMESPACE_GCO));
            if (href != null) {
                resource.addProperty(property, model.createResource(href));
            } else if (nilReason != null) {
                //TODO Is it correct to do nothing for a nil value?
            } else {
                throw new IllegalArgumentException("A GML property has one child element or a xlink:href attribute. " + element.getName());
            }
        } else if (numberOfChildElements == 1) {
            final Element entity = element.elementIterator().next();
            resource.addProperty(property, workOnEntity(model, entity));
        } else {
            throw new IllegalArgumentException("A GML property has one child element or a xlink:href attribute. " + element.getName());
        }

    }
}
