package de.landsh.opendata.csw2dcat;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.ResourceFactory;
import org.dom4j.Element;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Objects;

/**
 * Map ISO 19115 vocabulary to DCAT-AP compliant vocabulary.
 */
public class Mapping {

    /**
     * Formats in the list of the EU publications office
     * <a href="http://publications.europa.eu/resource/authority/file-type">http://publications.europa.eu/resource/authority/file-type</a>
     */
    public static final String[] ALLOWED_FORMATS;
    public static final String[] SERVICES;
    private static final Logger log = LoggerFactory.getLogger(Mapping.class);

    static {
        log.debug("Reading list of allowed file formats...");
        try {
            ALLOWED_FORMATS = IOUtils.readLines(Objects.requireNonNull(Mapping.class.getResourceAsStream("/allowed-formats.txt")), StandardCharsets.UTF_8).toArray(new String[0]);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        log.debug("Reading list of allowed service formats...");
        try {
            SERVICES = IOUtils.readLines(Objects.requireNonNull(Mapping.class.getResourceAsStream("/allowed-services.txt")), StandardCharsets.UTF_8).toArray(new String[0]);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public static String mapScopeCode(Element element) {
        if (element == null) {
            return null;
        }
        final String codeList = element.attributeValue("codeList");
        final String codeListValue = element.attributeValue("codeListValue");
        if ("https://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_ScopeCode".equals(codeList)
                || "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_ScopeCode".equals(codeList)
                || "MD_ScopeCode".equals(codeList)
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#MD_ScopeCode".equals(codeList)
                || StringUtils.startsWith(codeList, "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_ScopeCode_")
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/Codelist/ML_gmxCodelists.xml#MD_ScopeCode".equals(codeList)
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_ScopeCode".equals(codeList)
        ) {
            return "http://inspire.ec.europa.eu/metadata-codelist/ResourceType/" + codeListValue;
        } else {
            log.info("Unknown codeList {} for MD_ScopeCode", codeList);
            throw new IllegalArgumentException("Unknown codeList " + codeList + " for MD_ScopeCode");
        }
    }

    public static String mapCharacterSetCode(Element element) {
        if (element == null) {
            return null;
        }

        final String codeList = element.attributeValue("codeList");
        final String codeListValue = element.attributeValue("codeListValue");
        if ("https://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_CharacterSetCode".equals(codeList)
                || "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_CharacterSetCode".equals(codeList)
                || "MD_CharacterSetCode".equals(codeList)
                || StringUtils.startsWith(codeList, "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_CharacterSetCode_")
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/Codelist/ML_gmxCodelists.xml#MD_CharacterSetCode".equals(codeList)
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#MD_CharacterSetCode".equals(codeList)
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_CharacterSetCode".equals(codeList)
        ) {
            if ("utf8".equals(codeListValue)) {
                return "utf8";
            } else {
                return null;
            }
        } else {
            log.info("Unknown codeList {} for MD_CharacterSetCode, value {}", codeList, codeListValue);
            throw new IllegalArgumentException("Unknown codeList " + codeList + " for MD_CharacterSetCode, value " + codeListValue);
        }
    }

    public static String mapLanguageCode(Element element) {
        if (element == null) {
            return null;
        }

        final String codeList = element.attributeValue("codeList");
        final String codeListValue = element.attributeValue("codeListValue");
        if ("http://www.loc.gov/standards/iso639-2/".equals(codeList)) {
            if ("ger".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/language/DEU";
            } else {
                return null;
            }
        } else {
            log.info("Unknown codeList {} for LanguageCode", codeList);
            throw new IllegalArgumentException("Unknown codeList " + codeList + " for LanguageCode");
        }
    }

    public static String mapMaintenanceFrequencyCode(Element element) {
        if (element == null) {
            return null;
        }

        final String codeList = element.attributeValue("codeList");
        final String codeListValue = element.attributeValue("codeListValue");
        if ("https://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_MaintenanceFrequencyCode".equals(codeList)
                || "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_MaintenanceFrequencyCode".equals(codeList)
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/Codelist/ML_gmxCodelists.xml#MD_MaintenanceFrequencyCode".equals(codeList)
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#MD_MaintenanceFrequencyCode".equals(codeList)
                || "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_MaintenanceFrequencyCode".equals(codeList)
                || "MD_MaintenanceFrequencyCode".equals(codeList)) {
            if ("annually".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/ANNUAL";
            } else if ("asNeeded".equals(codeListValue) || "irregular".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/IRREG";
            } else if ("biannually".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/ANNUAL_2";
            } else if ("continual".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/CONT";
            } else if ("daily".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/DAILY";
            } else if ("fortnightly".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/BIWEEKLY";
            } else if ("monthly".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/MONTHLY";
            } else if ("notPlanned".equals(codeListValue)) {
                return "http://inspire.ec.europa.eu/metadata-codelist/MaintenanceFrequencyCode/notPlanned";
            } else if ("quarterly".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/QUARTERLY";
            } else if ("unknown".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/UNKNOWN";
            } else if ("weekly".equals(codeListValue)) {
                return "http://publications.europa.eu/resource/authority/frequency/WEEKLY";
            } else {
                if (StringUtils.isNotBlank(codeListValue)) {
                    log.warn("Unknown value {} for MD_MaintenanceFrequencyCode", codeListValue);
                }
                return null;
            }
        } else {
            log.info("Unknown codeList {} for MD_MaintenanceFrequencyCode", codeList);
            throw new IllegalArgumentException("Unknown codeList " + codeList + " for MD_MaintenanceFrequencyCode");
        }
    }

    public static Resource mapFormat(String format) {
        if (ArrayUtils.contains(ALLOWED_FORMATS, format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/" + format);
        } else if (ArrayUtils.contains(SERVICES, format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/" + format + "_SRVC");
        } else if ("download".equals(format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/WFS_SRVC");
        } else if ("view".equals(format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/WMS_SRVC");
        } else if ("Shapefiles".equals(format) || "Shape".equals(format) || "SHAPE".equals(format) || "ESRI-Shape".equals(format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/SHP");
        } else if ("TIFF 5 Group 4 Fax CCITT, TIFF-LZW (Summenlayer)".equals(format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/TIFF");
        } else if ("predefined ATOM".equals(format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/ATOM");
        } else if ("Excel".equals(format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/XLS");
        } else if ("GeoTIFF".equals(format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/GEOTIFF");
        } else if ("ARC/INFO-Coverage".equals(format)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/ARCINFO_COV");
        } else if (StringUtils.isEmpty(format)) {
            throw new IllegalArgumentException("Format must not be empty.");
        } else {
            log.info("{} is not an allowed format.", format);
            throw new IllegalArgumentException(format + " is not an allowed format.");
        }
    }

    /**
     * Map the content of a gmd:CI_OnlineResource/gmd:applicationProfile> element to a
     * <a href="http://publications.europa.eu/resource/authority/file-type">http://publications.europa.eu/resource/authority/file-type</a>
     * URI
     */
    public static Resource mapApplicationProfile(String applicationProfile) {
        if ("GeoTIFF".equals(applicationProfile)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/GEOTIFF");
        } else if ("PDF".equals(applicationProfile)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/PDF");
        } else if ("Excel".equals(applicationProfile)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/XLS");
        } else if ("Shapefiles".equals(applicationProfile)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/SHP");
        } else if ("JPEG".equals(applicationProfile)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/JPEG");
        } else if ("CSV".equals(applicationProfile)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/CSV");
        } else if ("HTML".equals(applicationProfile)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/HTML");
        } else if ("GML".equals(applicationProfile)) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/GML");
        } else {
            log.info("{} is not an allowed applicationProfile.", applicationProfile);
            throw new IllegalArgumentException(applicationProfile + " is not an allowed applicationProfile.");
        }
    }

    public static Resource mapHvdCategory(String germanName) {
        if ("Georaum".equals(germanName)) {
            return ResourceFactory.createResource("http://data.europa.eu/bna/c_ac64a52d");
        } else if ("Erdbeobachtung und Umwelt".equals(germanName)) {
            return ResourceFactory.createResource("http://data.europa.eu/bna/c_dd313021");
        } else if ("Erdbeobachtung".equals(germanName)) {
            return ResourceFactory.createResource("http://data.europa.eu/bna/c_dd313021");
        } else if ("Meteorologie".equals(germanName)) {
            return ResourceFactory.createResource("http://data.europa.eu/bna/c_164e0bf5");
        } else if ("Statistik".equals(germanName)) {
            return ResourceFactory.createResource("http://data.europa.eu/bna/c_e1da4e07");
        } else if ("Unternehmen und Eigentümerschaft von Unternehmen".equals(germanName)) {
            return ResourceFactory.createResource("http://data.europa.eu/bna/c_a9135398");
        } else if ("Mobilität".equals(germanName)) {
            return ResourceFactory.createResource("http://data.europa.eu/bna/c_b79e35eb");
        } else {
            log.info("{} is not an allowed HVD category.", germanName);
            throw new IllegalArgumentException(germanName + " is not an allowed HVD category.");
        }
    }
}
