package de.landsh.opendata.csw2dcat;

import de.landsh.opendata.csw2dcat.vocabulary.Hydra;
import io.undertow.Undertow;
import io.undertow.server.HttpHandler;
import io.undertow.server.HttpServerExchange;
import io.undertow.util.Headers;
import io.undertow.util.HttpString;
import io.undertow.util.Methods;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.apache.hc.core5.http.HttpStatus;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.ResourceFactory;
import org.apache.jena.vocabulary.DCAT;
import org.apache.jena.vocabulary.RDF;
import org.dom4j.Element;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.StringWriter;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;

public class Server implements HttpHandler {
    static final Map<Integer, BridgeSettings> bridgeSettings = new HashMap<>();
    private static final Logger log = LoggerFactory.getLogger(Server.class);
    public static int port = 8080;
    static String baseURL = "http://localhost:8080";

    static String dcatDatasetBaseURL = "https://opendata.schleswig-holstein.de/dataset/";

    private static String dumpCswResponse = null;
    private static boolean enabledDcat2IsoBridge = false;
    private static String ckanApiURL;
    private static Dcat2iso dcat2iso;
    /**
     * Name of a directory. If set all incoming and outgoing messages will be logged into files in this directory.
     */
    private static String verboseLogging;

    /**
     * Parse the environment settings for one bridge.
     */
    private static BridgeSettings parseEnvironmentSettingsForOneBridge(String prefix) {
        final String bridgeURL = System.getenv("BRIDGE" + prefix + "_URL");
        final String bridgeContributor = System.getenv("BRIDGE" + prefix + "_CONTRIBUTOR");
        final String bridgeIRI = System.getenv("BRIDGE" + prefix + "_IRI");

        if (StringUtils.isBlank(bridgeContributor)) {
            log.warn("BRIDGE" + prefix + "_CONTRIBUTOR environment variable is unset. Using example value.");
        }
        if (StringUtils.isBlank(bridgeIRI)) {
            log.warn("BRIDGE" + prefix + "_IRI environment variable is unset. Using example value.");
        }

        final BridgeSettings settings = new BridgeSettings();
        settings.cswURL = bridgeURL;
        settings.baseIRI = StringUtils.defaultString(bridgeIRI, "https://example.org/");
        settings.contributorId = StringUtils.defaultString(bridgeContributor, "https://example.org/contributor");

        settings.distributedSearch = BooleanUtils.toBoolean(System.getenv("BRIDGE" + prefix + "_DISTRIBUTED"));
        if (System.getenv("BRIDGE" + prefix + "_TYPE_EXCLUDE") != null) {
            String value = System.getenv("BRIDGE" + prefix + "_TYPE_EXCLUDE").replaceAll(" ", StringUtils.EMPTY);
            settings.typeExclude = value.split(",");
        }
        if (BooleanUtils.toBoolean(System.getenv("BRIDGE" + prefix + "_EXCLUDE_TILES"))) {
            if (settings.typeExclude == null) {
                settings.typeExclude = new String[]{"tile"};
            } else {
                settings.typeExclude = ArrayUtils.add(settings.typeExclude, "tile");
            }
        }
        if (System.getenv("BRIDGE" + prefix + "_TYPE_INCLUDE") != null) {
            String value = System.getenv("BRIDGE" + prefix + "_TYPE_INCLUDE").replaceAll(" ", StringUtils.EMPTY);
            settings.typeInclude = value.split(",");
        }
        settings.findWMSinInfo = BooleanUtils.toBoolean(System.getenv("BRIDGE" + prefix + "_FIND_WMS_IN_INFO"));
        settings.addMissingTemporalStart = BooleanUtils.toBoolean(System.getenv("BRIDGE" + prefix + "_FIX_MISSING_TEMPORAL"));
        settings.filterOpenData = BooleanUtils.toBoolean(System.getenv("BRIDGE" + prefix + "_FILTER_OPENDATA"));
        settings.authorization = StringUtils.defaultString(System.getenv("BRIDGE" + prefix + "_AUTHORIZATION"));

        if (System.getenv("BRIDGE" + prefix + "_SORT_RESULTS") != null) {
            settings.sortResults = BooleanUtils.toBoolean(System.getenv("BRIDGE" + prefix + "_SORT_RESULTS"));
        }

        if (StringUtils.isNotBlank(System.getenv("BRIDGE" + prefix + "_REMOVE_KEYWORDS"))) {
            settings.keywordRemove = Arrays.asList(System.getenv("BRIDGE" + prefix + "_REMOVE_KEYWORDS").split(","));
        }

        return settings;
    }

    static void readEnvironmentVariables() {
        if (StringUtils.isNotBlank(System.getenv("DCAT_BASE_URL"))) {
            enabledDcat2IsoBridge = true;
            log.info("Enabling DCAT2ISO bridge.");
            dcatDatasetBaseURL = System.getenv("DCAT_BASE_URL");
        }

        if (StringUtils.isNotBlank(System.getenv("CKAN_API_URL"))) {
            ckanApiURL = System.getenv("CKAN_API_URL");
        }

        if (StringUtils.isNotBlank(System.getenv("BRIDGE_BASE_URL"))) {
            baseURL = System.getenv("BRIDGE_BASE_URL");
        }

        if (StringUtils.isNotBlank(System.getenv("BRIDGE_PORT"))) {
            port = Integer.parseInt(System.getenv("BRIDGE_PORT"));
        }

        if (StringUtils.isNotBlank(System.getenv("BRIDGE_DUMP_CSW_RESPONSE"))) {
            dumpCswResponse = System.getenv("BRIDGE_DUMP_CSW_RESPONSE");
        }

        if (StringUtils.isNotBlank(System.getenv("BRIDGE_VERBOSE_LOGGING"))) {
            verboseLogging = System.getenv("BRIDGE_VERBOSE_LOGGING");

            // Check if the specified path is an existing direcotry.
            final File dir = new File(verboseLogging);
            if (dir.isDirectory()) {
                if( dir.canWrite()) {
                    log.info("Writing incoming and outgoing messages into {}", verboseLogging);
                } else {
                    log.warn("Cannot write in the directory for verbose logging {}. No files will be written", verboseLogging);
                    verboseLogging = null;
                }
            } else {
                log.warn("Location for verbose logging {} is not a directory. No files will be written.", verboseLogging);
                verboseLogging = null;
            }
        }

        if (StringUtils.isNotBlank(System.getenv("BRIDGE_URL"))) {
            // just a single CSW2DCAT bridge
            bridgeSettings.put(0, parseEnvironmentSettingsForOneBridge(StringUtils.EMPTY));
        } else if (StringUtils.isNotBlank(System.getenv("BRIDGE_1_URL"))) {
            // multiple CSW2DCAT bridges

            for (int i = 1; i <= 9; i++) {
                final String bridgeURL = System.getenv("BRIDGE_" + i + "_URL");

                if (StringUtils.isBlank(bridgeURL)) {
                    // the previously configured bridge was the last one
                    break;
                }

                bridgeSettings.put(i - 1, parseEnvironmentSettingsForOneBridge("_" + i));
            }
        } else {
            System.err.println("Missing environment variable BRIDGE_URL.");
            System.err.println("For example, you can start the program like this:");
            System.err.println("BRIDGE_URL=https://umweltportal.schleswig-holstein.de/csw  java -jar csw2dcat.jar");
            System.exit(1);
        }
    }

    public static void main(String[] args) throws IOException {
        readEnvironmentVariables();

        if (enabledDcat2IsoBridge) {
            dcat2iso = new Dcat2iso("ckan", baseURL, dcatDatasetBaseURL, ckanApiURL);
        }

        final Undertow server = Undertow.builder()
                .addHttpListener(port, "0.0.0.0", new Server()).build();
        server.start();

        System.out.println("\nRunning! Point your browser to " + baseURL + "/catalog.xml \n");

    }

    void addHydraPageCollection(Model model, int page, int itemsPerPage, int totalItems,
                                String catalogURL, Map<String, Set<String>> filters) {
        // add hydra:PagedCollection element

        final StringBuilder sb = new StringBuilder();
        for (String parameterName : filters.keySet()) {
            for (String parameterValue : filters.get(parameterName)) {
                sb.append('&');
                sb.append(parameterName);
                sb.append('=');
                sb.append(URLEncoder.encode(parameterValue, StandardCharsets.UTF_8));
            }
        }
        final String query = sb.toString();

        final Resource pagedCollection = model.createResource(catalogURL + "?page=" + page + query);
        pagedCollection.addProperty(RDF.type, Hydra.PagedCollection);
        pagedCollection.addLiteral(Hydra.itemsPerPage, ResourceFactory.createTypedLiteral(itemsPerPage));
        pagedCollection.addLiteral(Hydra.totalItems, ResourceFactory.createTypedLiteral(totalItems));
        pagedCollection.addProperty(Hydra.firstPage, model.createResource(catalogURL + "?page=1" + query));

        if (totalItems > page * itemsPerPage) {
            pagedCollection.addProperty(Hydra.nextPage, model.createResource(catalogURL + "?page=" + (page + 1) + query));
        }
        if (page > 1) {
            pagedCollection.addProperty(Hydra.previousPage, model.createResource(catalogURL + "?page=" + (page - 1) + query));
        }
        pagedCollection.addProperty(Hydra.lastPage, model.createResource(catalogURL + "?page=" + (totalItems / itemsPerPage + 1) + query));
    }

    @Override
    public void handleRequest(HttpServerExchange exchange) {
        final String requestURI = exchange.getRequestPath();
        if ("/catalog.xml".equals(requestURI)) {
            this.csw2dcat(exchange, 0);
        } else if (requestURI.matches("^/[0-9]+/catalog.xml$")) {
            int bridgeNumber = NumberUtils.toInt(StringUtils.substringBetween(requestURI, "/")) - 1;
            this.csw2dcat(exchange, bridgeNumber);
        } else if (enabledDcat2IsoBridge && "/iso/list.json".equals(requestURI)) {
            dcat2iso.list(exchange, "GDI-SH");
        } else if (enabledDcat2IsoBridge && StringUtils.startsWith(requestURI, "/iso/")) {
            dcat2iso.handleDcat2isoRequest(exchange);
        } else if ("/live".equals(requestURI)) {
            exchange.setStatusCode(HttpStatus.SC_OK);
            exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "text/plain");
            exchange.getResponseSender().send("ok");
        } else if ("/ready".equals(requestURI)) {
            exchange.setStatusCode(HttpStatus.SC_OK);
            exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "text/plain");
            exchange.getResponseSender().send("ok");
        }
    }


    /**
     * Handle CSW -> DCAT request
     */
    void csw2dcat(HttpServerExchange exchange, int bridgeNumber) {

        log.info("Using CSW2DCAT bridge #{}", bridgeNumber);

        if (!bridgeSettings.containsKey(bridgeNumber)) {
            exchange.setStatusCode(HttpStatus.SC_NOT_FOUND);
            exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "text/plain");
            exchange.getResponseSender().send("Not Found");
            return;
        }

        final BridgeSettings settings = bridgeSettings.get(bridgeNumber);

        if (Methods.HEAD.equals(exchange.getRequestMethod())) {
            exchange.setStatusCode(HttpStatus.SC_OK);
            exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "application/rdf+xml");
            exchange.getResponseHeaders().put(HttpString.tryFromString("Access-Control-Allow-Origin"), "*");
            return;
        }

        final Model model = ModelFactory.createDefaultModel();


        // Extract parameters from request
        final int page;
        if (exchange.getQueryParameters().containsKey("page")) {
            page = NumberUtils.toInt(exchange.getQueryParameters().get("page").getFirst());
            if (page == 0) {
                exchange.setStatusCode(HttpStatus.SC_BAD_REQUEST);
                exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "text/plain");
                exchange.getResponseSender().send("Invalid 'page' parameter");
                return;
            }
        } else {
            page = 1;
        }

        final Map<String, Set<String>> filters = new HashMap<>();
        if (exchange.getQueryParameters().containsKey("Subject")) {
            filters.put("Subject", new HashSet<>(exchange.getQueryParameters().get("Subject")));
        }
        if (exchange.getQueryParameters().containsKey("OrganisationName")) {
            filters.put("OrganisationName", new HashSet<>(exchange.getQueryParameters().get("OrganisationName")));
        }
        if (exchange.getQueryParameters().containsKey("Identifier")) {
            filters.put("Identifier", new HashSet<>(exchange.getQueryParameters().get("Identifier")));
        }

        final CswInterface cswInterface = new CswInterface(settings);
        if (dumpCswResponse != null) {
            log.info("Dumping CSW response to {}", dumpCswResponse);
            cswInterface.setDumpCswResponse(dumpCswResponse);
        }
        if (verboseLogging != null) {
            cswInterface.setVerboseLogging(verboseLogging);
        }

        // Prepare the converter
        final MDMetadata2Dataset mdMetadata2Dataset = new MDMetadata2Dataset(model,
                cswInterface,
                settings);

        // Prepare the resulting RDF document
        final Resource catalog = model.createResource();
        catalog.addProperty(RDF.type, DCAT.Catalog);

        int startPosition = (page - 1) * cswInterface.getRecordsPerPage() + 1;
        log.info("Getting records from CSW starting at position {}", startPosition);

        try {
            final CswInterface.GetRecordsStatus status = new CswInterface.GetRecordsStatus();
            int count = 0;
            for (Element element : cswInterface.getRecords(startPosition, status, filters)) {
                try {
                    final Resource dataset = mdMetadata2Dataset.convert(element);
                    if (dataset != null) {
                        catalog.addProperty(DCAT.dataset, dataset);
                    }
                } catch (Exception e) {
                    log.warn("Failed to convert element. {}: {}", e.getClass().getSimpleName(), e.getMessage());
                    if (log.isDebugEnabled()) {
                        final StringWriter sw = new StringWriter();
                        element.write(sw);
                        sw.close();
                        log.debug("Failed element: {}", sw);
                    }
                }
                count++;
            }
            log.info("Received {} records", count);

            final String catalogURL = baseURL + "/" + (bridgeNumber + 1) + "/catalog.xml";
            addHydraPageCollection(catalog.getModel(), page, cswInterface.getRecordsPerPage(),
                    status.getNumberOfRecordsMatched(), catalogURL, filters);

            CommandLine.setNsPrefixes(model);

            final StringWriter out = new StringWriter();
            model.write(out, "RDF/XML-ABBREV");
            out.close();

            if (verboseLogging != null) {
                // Write the response into  file
                final File dir = new File(verboseLogging);

                if (dir.isDirectory()) {
                    final File file = new File(dir, "out_" + DateTimeFormatter.ISO_LOCAL_DATE_TIME.format(LocalDateTime.now()) + ".rdf");
                    FileOutputStream fos = new FileOutputStream(file);
                    fos.write(out.toString().getBytes(StandardCharsets.UTF_8));
                    fos.close();
                } else {
                    log.warn("Location for verbose logging {} is not a directory. No file written.", verboseLogging);
                }

            }

            exchange.setStatusCode(HttpStatus.SC_OK);
            exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "application/rdf+xml");
            exchange.getResponseHeaders().put(HttpString.tryFromString("Access-Control-Allow-Origin"), "*");
            exchange.getResponseSender().send(out.toString());


            log.debug("Done.");
        } catch (Exception e) {
            e.printStackTrace();
            exchange.setStatusCode(HttpStatus.SC_INTERNAL_SERVER_ERROR);
            exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, "text/plain");
            exchange.getResponseSender().send("Internal Server Error");
        }
    }

}
