package de.landsh.opendata.csw2dcat;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.hc.client5.http.classic.methods.HttpPost;
import org.apache.hc.client5.http.impl.classic.CloseableHttpClient;
import org.apache.hc.client5.http.impl.classic.CloseableHttpResponse;
import org.apache.hc.core5.http.HttpHeaders;
import org.apache.hc.core5.http.ProtocolException;
import org.apache.hc.core5.http.io.entity.StringEntity;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.io.SAXReader;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;
import org.mockserver.client.MockServerClient;
import org.mockserver.integration.ClientAndServer;
import org.mockserver.model.Header;
import org.mockserver.model.HttpStatusCode;

import java.io.IOException;
import java.io.StringReader;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.concurrent.TimeUnit;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockserver.integration.ClientAndServer.startClientAndServer;
import static org.mockserver.matchers.Times.exactly;
import static org.mockserver.model.HttpRequest.request;
import static org.mockserver.model.HttpResponse.response;

class CswInterfaceTest {

    private static ClientAndServer mockServer;
    SAXReader saxReader = new SAXReader();
    private CswInterface cswInterface;

    @BeforeAll
    public static void startServer() {
        mockServer = startClientAndServer(1080);
    }

    @AfterAll
    public static void stopServer() {
        mockServer.stop(true);
        while (!mockServer.hasStopped(3, 100L, TimeUnit.MILLISECONDS)) {
        }
    }

    private static String wrapXML(String result) {
        return "<xml" +
                " xmlns:ogc=\"http://www.opengis.net/ogc\"" +
                " xmlns:apiso=\"http://www.opengis.net/cat/apiso/1.0\">" + result + "</xml>";
    }

    @BeforeEach
    public void setUp() {

        cswInterface = new CswInterface("http://localhost:" + mockServer.getPort() + "/api");

    }

    @Test
    public void findOperatesOn() throws IOException, DocumentException {
        String data = IOUtils.toString(Objects.requireNonNull(getClass().getResourceAsStream("/csw_response_1.xml")), StandardCharsets.UTF_8);

        new MockServerClient("127.0.0.1", mockServer.getPort())
                .when(
                        request()
                                .withMethod("POST")
                                .withPath("/api"),
                        exactly(1))
                .respond(
                        response()
                                .withStatusCode(HttpStatusCode.OK_200.code())
                                .withHeaders(
                                        new Header("Content-Type", "text/xml; charset=utf-8"))
                                .withBody(data)
                );


        List<Element> result = cswInterface.findOperatesOn("7164d86e-8c55-4e08-ab1f-601b3002d1d1");
        assertNotNull(result);
        assertEquals(2, result.size());
    }

    @Test
    void buildFilterQuery_empty() {
        String result = cswInterface.buildFilterQuery(new HashMap<>());

        assertEquals(StringUtils.EMPTY, result);
    }

    @Test
    void buildFilterQuery() throws DocumentException {
        Map<String, Set<String>> filters = new HashMap<>();
        filters.put("Subject", Set.of("foo", "bar"));
        filters.put("OrganisationName", Set.of("ZIT"));

        String result = cswInterface.buildFilterQuery(filters);

        Element filter = saxReader.read(new StringReader(wrapXML(result))).getRootElement().element("And");
        List<Element> properties = filter.elements("PropertyIsEqualTo");
        assertEquals(3, properties.size());

        Set<String> propertyFilter = new HashSet<>();
        for (Element property : properties) {
            propertyFilter.add(property.elementText("PropertyName") + " = " + property.elementText("Literal"));
        }

        assertTrue(propertyFilter.contains("apiso:OrganisationName = ZIT"));
        assertTrue(propertyFilter.contains("apiso:Subject = foo"));
        assertTrue(propertyFilter.contains("apiso:Subject = bar"));
    }

    @Test
    void buildFilterQuery_filterOpenData() throws DocumentException {

        Map<String, Set<String>> filters = new HashMap<>();
        filters.put("Subject", new HashSet<>(Set.of("foo", "bar")));
        filters.put("OrganisationName", Set.of("ZIT"));

        cswInterface.setFilterOpendata(true);
        String result = cswInterface.buildFilterQuery(filters);

        Element filter = saxReader.read(new StringReader(wrapXML(result))).getRootElement().element("And");
        List<Element> properties = filter.elements("PropertyIsEqualTo");
        assertEquals(4, properties.size());

        Set<String> propertyFilter = new HashSet<>();
        for (Element property : properties) {
            propertyFilter.add(property.elementText("PropertyName") + " = " + property.elementText("Literal"));
        }

        assertTrue(propertyFilter.contains("apiso:OrganisationName = ZIT"));
        assertTrue(propertyFilter.contains("apiso:Subject = foo"));
        assertTrue(propertyFilter.contains("apiso:Subject = bar"));
        assertTrue(propertyFilter.contains("apiso:Subject = opendata"));
    }

    @Test
    void buildFilterQuery_empty_filterOpenData() throws DocumentException {
        cswInterface.setFilterOpendata(true);
        String result = cswInterface.buildFilterQuery(new HashMap<>());

        Element filter = saxReader.read(new StringReader(wrapXML(result))).getRootElement().element("And");
        Element property = filter.element("PropertyIsEqualTo");
        assertNotNull(property);
        assertEquals("apiso:Subject", property.elementText("PropertyName"));
        assertEquals("opendata", property.elementText("Literal"));

    }

    /**
     * Prüfe, dass die XML-Anfrage für die Suche nach gekoppelten Diensten korrekt ist.
     */
    @Test
    void findOperatesOn_checkXML() throws DocumentException, IOException, ProtocolException {
        final CloseableHttpClient client = Mockito.mock(CloseableHttpClient.class);
        final ArgumentCaptor<HttpPost> argumentCaptor = ArgumentCaptor.forClass(HttpPost.class);
        final CloseableHttpResponse mockResponse = Mockito.mock(CloseableHttpResponse.class);
        StringEntity entity = new StringEntity("<xml/>", StandardCharsets.UTF_8);
        Mockito.when(mockResponse.getEntity()).thenReturn(entity);

        Mockito.when(client.execute(argumentCaptor.capture())).thenReturn(mockResponse);

        cswInterface.setHttpClient(client);
        cswInterface.findOperatesOn("5bf8d8c6-01bf-465b-98de-a0f32f3d739e");

        HttpPost request = argumentCaptor.getValue();

        assertEquals("text/xml", request.getHeader(HttpHeaders.CONTENT_TYPE).getValue());
        assertEquals("/api", request.getRequestUri());

        Element xml = saxReader.read(request.getEntity().getContent()).getRootElement();
        assertEquals("GetRecords", xml.getName());
        assertNotNull(xml.element("Query"));
        assertEquals("full", xml.element("Query").elementText("ElementSetName"));
        assertNotNull(xml.element("Query").element("Constraint"));
        Element filter = xml.element("Query").element("Constraint").element("Filter");
        assertNotNull(filter );
        assertNotNull( filter.element("PropertyIsEqualTo"));
        assertEquals("operatesOn", filter.element("PropertyIsEqualTo").elementText("PropertyName"));
        assertEquals("5bf8d8c6-01bf-465b-98de-a0f32f3d739e", filter.element("PropertyIsEqualTo").elementText("Literal"));
    }

    /**
     * Ist <code>filterOpendata</code> gesetzt, dann wird ein And-Element eingefügt und darin ein Filter
     * Subject=opendata gesetzt.
     */
    @Test
    void findOperatesOn_filterOpendata() throws DocumentException, IOException, ProtocolException {
        // mock methods and responses
        final CloseableHttpClient client = Mockito.mock(CloseableHttpClient.class);
        final ArgumentCaptor<HttpPost> argumentCaptor = ArgumentCaptor.forClass(HttpPost.class);
        final CloseableHttpResponse mockResponse = Mockito.mock(CloseableHttpResponse.class);
        StringEntity entity = new StringEntity("<xml/>", StandardCharsets.UTF_8);
        Mockito.when(mockResponse.getEntity()).thenReturn(entity);

        Mockito.when(client.execute(argumentCaptor.capture())).thenReturn(mockResponse);
        cswInterface.setHttpClient(client);

        // invoke method
        cswInterface.setFilterOpendata(true);
        cswInterface.findOperatesOn("d8fee8c2-4385-46df-ab5e-8a372ac5a04d");

        // check results
        HttpPost request = argumentCaptor.getValue();
        Element xml = saxReader.read(request.getEntity().getContent()).getRootElement();
        assertEquals("GetRecords", xml.getName());
        assertNotNull(xml.element("Query"));
        assertEquals("full", xml.element("Query").elementText("ElementSetName"));
        assertNotNull(xml.element("Query").element("Constraint"));
        Element filter = xml.element("Query").element("Constraint").element("Filter");
        assertNotNull(filter);

        Element elementAnd =   filter.element("And");
        assertNotNull( elementAnd);
        assertEquals( 2, elementAnd.elements("PropertyIsEqualTo").size());

        Map<String,String> propertyIsEqualTo = new HashMap<>();
        for (Element e : elementAnd.elements("PropertyIsEqualTo")) {
           propertyIsEqualTo.put(e.elementText("PropertyName"), e.elementText("Literal"));
        }

        assertEquals("d8fee8c2-4385-46df-ab5e-8a372ac5a04d", propertyIsEqualTo.get("operatesOn"));
        assertEquals("opendata", propertyIsEqualTo.get("apiso:Subject"));
    }

    @Test
    void buildFilterQuery_TypeExclude_One() throws DocumentException {
        cswInterface.setTypeExclude(new String[]{"tile"});
        String result = cswInterface.buildFilterQuery(new HashMap<>());
        Element xml = saxReader.read(new StringReader(wrapXML(result))).getRootElement();

        Element elementNot = xml.element("Not");
        assertNotNull(elementNot);
        assertEquals("apiso:Type", elementNot.element("PropertyIsEqualTo").elementText("PropertyName"));
        assertEquals("tile", elementNot.element("PropertyIsEqualTo").elementText("Literal"));
    }

    @Test
    void buildFilterQuery_TypeExclude_Multiple() throws DocumentException {
        cswInterface.setTypeExclude(new String[]{"tile", "application"});
        String result = cswInterface.buildFilterQuery(new HashMap<>());
        Element xml = saxReader.read(new StringReader(wrapXML(result))).getRootElement();

        Element andElement = xml.element("And");
        assertNotNull(andElement);

        List<Element> notElements = andElement.elements("Not");
        assertEquals("apiso:Type", notElements.get(0).element("PropertyIsEqualTo").elementText("PropertyName"));
        assertEquals("tile", notElements.get(0).element("PropertyIsEqualTo").elementText("Literal"));

        assertEquals("apiso:Type", notElements.get(1).element("PropertyIsEqualTo").elementText("PropertyName"));
        assertEquals("application", notElements.get(1).element("PropertyIsEqualTo").elementText("Literal"));
    }
}
