package de.landsh.opendata.csw2dcat;

import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Resource;
import org.dom4j.Element;
import org.junit.jupiter.api.Test;

import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;

class Dataset2MDMetadataTest {

    Dataset2MDMetadata service = new Dataset2MDMetadata();

    @Test
    void convert() throws IOException {
        final Model model = ModelFactory.createDefaultModel();
        model.read(getClass().getResourceAsStream("/dcat/stadtteile-kiel.rdf"), null);

        Resource dataset = model.getResource("https://opendata.schleswig-holstein.de/dataset/cf1497ae-e656-4811-b560-b8c71838203a");

        Element result = service.convert(dataset);

        //Writer writer = new FileWriter("/tmp/stadtteile-kiel.xml");
        //result.write(writer);
        //writer.flush();

        assertNotNull(result);
        assertEquals("MD_Metadata", result.getName());

        assertEquals("cf1497ae-e656-4811-b560-b8c71838203a", result.element("fileIdentifier").elementText("CharacterString"));
        assertEquals("http://www.loc.gov/standards/iso639-2/", result.element("language").element("LanguageCode").attributeValue("codeList"));
        assertEquals("ger", result.element("language").element("LanguageCode").attributeValue("codeListValue"));
    }

    /**
     * This dataset was read from SH-MIS, converted to DCAT-AP.de imported into CKAN and exported to DCAT-AP.de again.
     */
    @Test
    void convert2() throws IOException {
        final Model model = ModelFactory.createDefaultModel();
        model.read(getClass().getResourceAsStream("/dcat/b-plan-b-131.rdf"), null);

        Resource dataset = model.getResource("http://registry.gdi-de.org/id/de.sh/8547251b-7832-4426-a21a-8099bfaeffa5");

        Element result = service.convert(dataset);

        //Writer writer = new FileWriter("/tmp/bplan.xml");
        //result.write(writer);
        //writer.flush();

        assertNotNull(result);
        assertEquals("MD_Metadata", result.getName());
        assertNotNull(result.element("contact").element("CI_ResponsibleParty"));
        assertEquals("pointOfContact", result.element("contact").element("CI_ResponsibleParty").element("role").elementText("CI_RoleCode"));

        assertEquals("pointOfContact", result.element("contact")
                .element("CI_ResponsibleParty")
                .element("role")
                .element("CI_RoleCode")
                .attributeValue("codeListValue"));
        assertEquals("2022-06-21", result.element("dateStamp").elementText("Date"));

        assertNotNull(result.element("distributionInfo"));
        assertEquals(4, result.element("distributionInfo").element("MD_Distribution").elements("distributionFormat").size());
        assertEquals(7, result.element("distributionInfo").element("MD_Distribution").elements("transferOptions").size());

        Element MD_DataIdentification = result.element("identificationInfo").element("MD_DataIdentification");
        Element CI_Citation = MD_DataIdentification.element("citation").element("CI_Citation");
        assertEquals("B-Plan - B 131", CI_Citation.element("title").elementText("CharacterString"));
        assertEquals("2017-08-29", CI_Citation.selectSingleNode("gmd:date/gmd:CI_Date/gmd:date/gco:Date/text()").getText());
        assertNotNull(CI_Citation.selectSingleNode("gmd:date/gmd:CI_Date/gmd:dateType/gmd:CI_DateTypeCode[@codeListValue='publication']"));
        assertEquals("http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/ML_gmxCodelists.xml#CI_DateTypeCode",
                CI_Citation
                        .element("date").element("CI_Date").element("dateType").element("CI_DateTypeCode")
                        .attributeValue("codeList"));

        Element EX_TemporalExtent = (Element) MD_DataIdentification.selectSingleNode("gmd:extent/gmd:EX_Extent/gmd:temporalElement/gmd:EX_TemporalExtent");
        assertNotNull(EX_TemporalExtent);
        assertEquals("2017-08-29", EX_TemporalExtent.selectSingleNode("gmd:extent/gml:TimePeriod/gml:beginPosition/text()").getText());
        assertEquals("unknown", EX_TemporalExtent.selectSingleNode("gmd:extent/gml:TimePeriod/gml:endPosition/@indeterminatePosition").getText());

        Element EX_GeographicBoundingBox = (Element) MD_DataIdentification.selectSingleNode("gmd:extent/gmd:EX_Extent/gmd:geographicElement/gmd:EX_GeographicBoundingBox");
        assertNotNull(EX_GeographicBoundingBox);
        assertEquals(5.81, Double.parseDouble(EX_GeographicBoundingBox.element("westBoundLongitude").elementText("Decimal")));
        assertEquals(15.77, Double.parseDouble(EX_GeographicBoundingBox.element("eastBoundLongitude").elementText("Decimal")));
        assertEquals(54.76, Double.parseDouble(EX_GeographicBoundingBox.element("northBoundLatitude").elementText("Decimal")));
        assertEquals(47.26, Double.parseDouble(EX_GeographicBoundingBox.element("southBoundLatitude").elementText("Decimal")));
    }

    @Test
    void licenseAsText_DLby2() {
        String result = service.licenseAsText("http://dcat-ap.de/def/licenses/dl-by-de/2.0", "ACME GmbH");
        assertEquals("Für die Nutzung der Daten ist die Datenlizenz Deutschland Namensnennung 2.0 anzuwenden. " +
                "Die Lizenz ist über https://www.govdata.de/dl-de/by-2-0 abrufbar. " +
                "Der Quellenvermerk gemäß (2) der Lizenz lautet: ACME GmbH.", result);
    }

    @Test
    void licenseAsJSON_DLby2() {
        String result = service.licenseAsJSON("http://dcat-ap.de/def/licenses/dl-by-de/2.0", "ACME GmbH");
        assertEquals("{\"id\":\"dl-by-de/2.0\"," +
                "\"name\":\"Datenlizenz Deutschland Namensnennung 2.0\"," +
                "\"url\":\"https://www.govdata.de/dl-de/by-2-0\"," +
                "\"quelle\":\"ACME GmbH\"}", result);
    }

    @Test
    void convert3() throws IOException {
        final Model model = ModelFactory.createDefaultModel();
        model.read(getClass().getResourceAsStream("/dcat/alkis-schleswig-holstein-ohne-eigentumerangaben.rdf"), null);

        Resource dataset = model.getResource("http://registry.gdi-de.org/id/de.sh/01e24e0a-fda2-4a0f-ac71-bd8011749b11");

        Element result = service.convert(dataset);
        assertNotNull(result);
    }

    /**
     * The format of a distribution is specified as a literal.
     */
    @Test
    void convert_formatLiteral() throws IOException {
        final Model model = ModelFactory.createDefaultModel();
        model.read(getClass().getResourceAsStream("/dcat/camping-im-kreis-pinneberg.rdf"), null);

        Resource dataset = model.getResource("https://opendata.schleswig-holstein.de/dataset/64da92fe-9fbc-4e04-8989-5436066d43f9");

        Element result = service.convert(dataset);
        assertNotNull(result);

        Writer writer = new FileWriter("/tmp/iso.xml");
        result.write(writer);
        writer.flush();
    }
}
