package de.landsh.opendata.csw2dcat;

import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.QName;
import org.junit.jupiter.api.Test;

import java.util.Map;

import static org.junit.jupiter.api.Assertions.*;

class MappingTest {

    private static Element createGmdElement(String localName, String codeList, String codeListValue, String content) {
        Element element = DocumentHelper.createElement(QName.get(localName, "gmd", "http://www.isotc211.org/2005/gmd"));
        element.addAttribute("codeList", codeList);
        element.addAttribute("codeListValue", codeListValue);
        if (content != null) {
            element.setText(content);
        }
        return element;
    }

    @Test
    void mapScopeCode() {
        assertNull(Mapping.mapScopeCode(null));

        Element element = createGmdElement(
                "MD_ScopeCode",
                "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_ScopeCode",
                "service",
                "service");

        String result = Mapping.mapScopeCode(element);

        assertEquals("http://inspire.ec.europa.eu/metadata-codelist/ResourceType/service", result);
    }

    /**
     * Some items contain the values also in the <code>codeList</code> attribute, e.g.
     * &lt;gmd:MD_ScopeCode codeList="http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_ScopeCode_dataset" codeListValue="dataset">dataset&lt;/gmd:MD_ScopeCode>
     */
    @Test
    void mapScopeCode_valueInCodeList() {
        Element element = createGmdElement(
                "MD_ScopeCode",
                "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_ScopeCode_dataset",
                "dataset",
                "dataset");


        String result = Mapping.mapScopeCode(element);

        assertEquals("http://inspire.ec.europa.eu/metadata-codelist/ResourceType/dataset", result);
    }

    @Test
    void mapCharacterSetCode() {
        assertNull(Mapping.mapCharacterSetCode(null));

        Element element = createGmdElement(
                "MD_CharacterSetCode",
                "https://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_CharacterSetCode",
                "utf8",
                "utf8");

        String result = Mapping.mapCharacterSetCode(element);

        assertEquals("utf8", result);
    }

    @Test
    void mapCharacterSetCode_valueInCodeList() {
        Element element = createGmdElement(
                "MD_CharacterSetCode",
                "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_CharacterSetCode_utf8",
                "utf8",
                "utf8");

        String result = Mapping.mapCharacterSetCode(element);

        assertEquals("utf8", result);
    }

    @Test
    void mapFormat() {
        assertEquals("http://publications.europa.eu/resource/authority/file-type/WFS_SRVC", Mapping.mapFormat("WFS").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/WFS_SRVC", Mapping.mapFormat("download").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/WMS_SRVC", Mapping.mapFormat("WMS").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/WMS_SRVC", Mapping.mapFormat("view").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/SHP", Mapping.mapFormat("Shapefiles").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/SHP", Mapping.mapFormat("Shape").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/TIFF", Mapping.mapFormat("TIFF 5 Group 4 Fax CCITT, TIFF-LZW (Summenlayer)").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/ATOM", Mapping.mapFormat("predefined ATOM").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/ARCINFO_COV", Mapping.mapFormat("ARC/INFO-Coverage").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/GEOTIFF", Mapping.mapFormat("GeoTIFF").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/XLS", Mapping.mapFormat("Excel").getURI());

        assertThrows(IllegalArgumentException.class, () -> Mapping.mapFormat(null));
        assertThrows(IllegalArgumentException.class, () -> Mapping.mapFormat(""));
        assertThrows(IllegalArgumentException.class, () -> Mapping.mapFormat("foobar"));
    }

    @Test
    void mapMaintenanceFrequencyCode() {
        assertNull(Mapping.mapMaintenanceFrequencyCode(null));

        String codeList = "http://standards.iso.org/iso/19139/resources/gmxCodelists.xml#MD_MaintenanceFrequencyCode";
        assertEquals("http://publications.europa.eu/resource/authority/frequency/ANNUAL",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "annually", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/IRREG",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "asNeeded", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/ANNUAL_2",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "biannually", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/CONT",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "continual", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/DAILY",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "daily", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/IRREG",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "irregular", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/MONTHLY",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "monthly", null)));
        assertEquals("http://inspire.ec.europa.eu/metadata-codelist/MaintenanceFrequencyCode/notPlanned",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "notPlanned", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/QUARTERLY",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "quarterly", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/UNKNOWN",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "unknown", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/WEEKLY",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "weekly", null)));
        assertEquals("http://publications.europa.eu/resource/authority/frequency/BIWEEKLY",
                Mapping.mapMaintenanceFrequencyCode(createGmdElement("MD_MaintenanceFrequencyCode",
                        codeList, "fortnightly", null)));
    }

    @Test
    void mapApplicationProfile() {
        assertEquals("http://publications.europa.eu/resource/authority/file-type/GEOTIFF",
                Mapping.mapApplicationProfile("GeoTIFF").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/PDF",
                Mapping.mapApplicationProfile("PDF").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/SHP",
                Mapping.mapApplicationProfile("Shapefiles").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/JPEG",
                Mapping.mapApplicationProfile("JPEG").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/CSV",
                Mapping.mapApplicationProfile("CSV").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/HTML",
                Mapping.mapApplicationProfile("HTML").getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/XLS",
                Mapping.mapApplicationProfile("Excel").getURI());

        assertThrows(IllegalArgumentException.class, () -> Mapping.mapApplicationProfile("foobar"));
    }
}
